/*
	DooM MaP StaTistics, by Frans P. de Vries.

Derived from:

	DooM PostScript Maps Utility, by Frans P. de Vries.

And thus from:

	Doom Editor Utility, by Brendon Wyber and Raphaël Quinet.

	You are allowed to use any parts of this code in another program, as
	long as you give credits to the authors in the documentation and in
	the program itself.  Read the file README for more information.

	This program comes with absolutely no warranty.

	DMMPST.C - Main program routines
*/


/* the includes */
#include "dmmpst.h"


/* global variables */
BCINT GameVersion = 0x00; /* which game and version? */
                          /* if you change this, bad things will happen to you... */
                 /* 0x00 = Shareware / Demo
                    0x01 = Registered
                    0x02 = Commercial (II/Final, Strife, Hexen)
                    0x04 = Ultimate / Shadow of the Serpent Riders
                    0x08 = Deathkings of the Dark Citadel
                   +0x00 = Doom
                   +0x10 = Heretic
                   +0x20 = Strife
                   +0x40 = Hexen
                  */
char *MainWad     = "DOOM.WAD"; /* name of the main WAD file */
char **PatchWads  = NULL;	 /* list of patch WAD files */
char *UserLvlNm   = NULL;  /* user defined level name */
char *WkTpPath    = NULL;  /* path to input DoomWiki templates */
FILE *WkFile      = NULL;  /* the output DoomWiki file */
Bool Verbose      = FALSE; /* verbose statistics? */
Bool WSkeleton    = FALSE; /* skeleton wiki? */
Bool WStatistics  = FALSE; /* statistics wiki? */
Bool WMapdata     = FALSE; /* map data wiki? */
Bool WThings      = FALSE; /* things wiki? */
Bool WBoomMP      = FALSE; /* Boom multiplayer in things wiki? */


/* local variables */
char *WkFlName   = NULL; /* name of the output DoomWiki file */
BCINT Episode    = 0;    /* the episode number */
BCINT Mission    = 0;    /* the mission number */

OptDesc options[] =    /* description of the command line options */
{
/* short/long names  type  message if true/changed  message if false  where to store value */
	{ "W",  "MAIN",      OPT_STRING,    "Main WAD file",          NULL, &MainWad     },
	{ "PW", "PWAD",      OPT_STRINGACC, "Patch WAD file",         NULL, &PatchWads   },
	{ "",   "FILE",      OPT_STRINGLIST,"Patch WAD file",         NULL, &PatchWads   },
	{ "N",  "NAME",      OPT_STRING,    "User level name",        NULL, &UserLvlNm   },
	{ "E",  "EPISODE",   OPT_INTEGER,   "Episode number",         NULL, &Episode     },
	{ "M",  "MISSION",   OPT_INTEGER,   "Mission number",         NULL, &Mission     },
	{ "I",  "WIKITEMPL", OPT_STRING,    "DoomWiki templates dir", NULL, &WkTpPath    },
	{ "O",  "WIKIFILE",  OPT_STRING,    "DoomWiki output file",   NULL, &WkFlName    },
	{ "V",  "VERBOSE",   OPT_BOOLEAN,   "Verbose statistics",     NULL, &Verbose     },
	{ "K",  "SKELETON",  OPT_BOOLEAN,   "Skeleton wiki output",   NULL, &WSkeleton   },
	{ "S",  "STATISTICS",OPT_BOOLEAN,   "Statistics wiki output", NULL, &WStatistics },
	{ "D",  "MAPDATA",   OPT_BOOLEAN,   "Map data wiki output",   NULL, &WMapdata    },
	{ "T",  "THINGS",    OPT_BOOLEAN,   "Things wiki output",     NULL, &WThings     },
	{ "B",  "BOOMMP",    OPT_BOOLEAN,   "Boom multiplayer things",NULL, &WBoomMP     },
	{ NULL, NULL,        OPT_END,       NULL,                     NULL, NULL         }
};


/*
	the main program
*/
int main( int argc, char *argv[])
{
	BCINT i;

	Credits( stdout);
	argv++;
	argc--;
	/* read command line options */
	ParseCommandLineOptions( argc, argv);

	/* load the Wad files */
	OpenMainWad( MainWad);
	if (PatchWads)
		while (PatchWads[0])
		{
			OpenPatchWad( PatchWads[0]);
			PatchWads++;
		}
	/* sanity check */
	CloseUnusedWadFiles();

	/* all systems go! */
	MainLoop();
	/* that's all, folks! */
	CloseWadFiles();
	exit( 0);
}


/*
	append a string to a null-terminated string list
*/
void AppendItemToList( char ***list, char *item)
{
	BCINT i = 0;

	if (*list)
	{
		/* count the number of elements in the list (last = null) */
		while ((*list)[i])
			i++;
		/* expand the list */
		*list = (char **) ResizeMemory( *list, (i + 2) * sizeof( char **));
	}
	else
		/* create a new list */
		*list = (char **) GetMemory( 2 * sizeof( char **));

	/* append the new element */
	(*list)[i] = item;
	(*list)[i + 1] = NULL;
}


/*
	handle command line options
*/
void ParseCommandLineOptions( int argc, char *argv[])
{
	BCINT optnum;

	while (argc > 0)
	{
		if (argv[0][0] != '-' && argv[0][0] != '+')
			ProgError( "options must start with '-' or '+'");
		strupr( argv[0]);
		if (!strcmp( argv[0], "-?") || !strcmp( argv[0], "-H") ||
		    !strcmp( argv[0], "-HELP"))
		{
			Usage( stdout);
			exit( 0);
		}
		for (optnum = 0; options[optnum].opt_type != OPT_END; optnum++)
		{
			if (!strcmp( &(argv[0][1]), options[optnum].short_name) ||
			    !strcmp( &(argv[0][1]), options[optnum].long_name))
			{
				switch (options[optnum].opt_type)
				{
				case OPT_BOOLEAN:
					if (argv[0][0] == '+') /* '+/-' reversed for print flags */
					{
						*((Bool *) (options[optnum].data_ptr)) = TRUE;
						if (options[optnum].msg_if_true)
							printf( "%s.\n", options[optnum].msg_if_true);
					}
					else
					{
						*((Bool *) (options[optnum].data_ptr)) = FALSE;
						if (options[optnum].msg_if_false)
							printf( "%s.\n", options[optnum].msg_if_false);
					}
					break;
				case OPT_INTEGER:
					if (argc <= 1)
						ProgError( "missing argument after \"%s\"", argv[0]);
					argv++;
					argc--;
					*((BCINT *) (options[optnum].data_ptr)) = atoi( argv[0]);
					if (options[optnum].msg_if_true)
						printf( "%s: %d.\n", options[optnum].msg_if_true, atoi( argv[0]));
					break;
				case OPT_STRING:
					if (argc <= 1)
						ProgError( "missing argument after \"%s\"", argv[0]);
					argv++;
					argc--;
					*((char **) (options[optnum].data_ptr)) = argv[0];
					if (options[optnum].msg_if_true)
						printf( "%s: %s.\n", options[optnum].msg_if_true, argv[0]);
					break;
				case OPT_STRINGACC:
					if (argc <= 1)
						ProgError( "missing argument after \"%s\"", argv[0]);
					argv++;
					argc--;
					AppendItemToList( (char ***) options[optnum].data_ptr, argv[0]);
					if (options[optnum].msg_if_true)
						printf( "%s: %s.\n", options[optnum].msg_if_true, argv[0]);
					break;
				case OPT_STRINGLIST:
					if (argc <= 1)
						ProgError( "missing argument after \"%s\"", argv[0]);
					while (argc > 1 && argv[1][0] != '-' && argv[1][0] != '+')
					{
						argv++;
						argc--;
						AppendItemToList( (char ***) options[optnum].data_ptr, argv[0]);
						if (options[optnum].msg_if_true)
							printf( "%s: %s.\n", options[optnum].msg_if_true, argv[0]);
					}
					break;
				default:
					ProgError( "unknown option type (BUG!)");
				}
				break;
			}
		}
		if (options[optnum].opt_type == OPT_END)
			ProgError( "invalid argument: \"%s\"", argv[0]);
		argv++;
		argc--;
	}
}


/*
	output the program usage to the specified file
*/
void Usage( FILE *where)
{
	fprintf( where, "Usage:\n");
	fprintf( where, "dmmpst [-w <main_wad_file>] [-pw <pwad_file>] [-file <pwad_files>...]\n");
	fprintf( where, "       [+v] [+k|s|d|t] [+b] [-n <name>] -e <episode> -m <mission>\n");
	fprintf( where, "       [-i <DoomWiki_templates_dir>] [-o <DoomWiki_file>]\n");
	fprintf( where, "   -w    Gives the name of the main WAD file (also -main); default is DOOM.WAD\n");
	fprintf( where, "   -pw   To add one patch WAD file to be loaded; may be repeated (also -pwad)\n");
	fprintf( where, "   -file To add a list of patch WAD files to be loaded\n");
	fprintf( where, "   +v    Display verbose statistics (also +verbose)\n");
	fprintf( where, "   +k    Generate complete skeleton to output file (also +skeleton)\n");
	fprintf( where, "   +s    Generate statistics section to output file (also +statistics)\n");
	fprintf( where, "   +d    Generate only map data template to output file (also +mapdata)\n");
	fprintf( where, "   +t    Generate only things template to output file (also +things)\n");
	fprintf( where, "   +b    Count Boom MP things instead of classic MP (also +boommp)\n");
	fprintf( where, "   -n    Defines the user level name (also -name)\n");
	fprintf( where, "   -e    Gives the episode number (also -episode); required\n");
	fprintf( where, "   -m    Gives the mission number (also -mission); required\n");
	fprintf( where, "   -i    Gives the path to input DoomWiki templates (also -wikitempl);\n");
	fprintf( where, "         if templates path is missing, uses current directory\n");
	fprintf( where, "   -o    Gives the name of the output DoomWiki file (also -wikifile);\n");
	fprintf( where, "         if output file is missing, [+k|s|d|t] are ignored\n");
	fprintf( where, "Put a '+' instead of a '-' before boolean options to reverse their effect\n");
}


/*
	output the credits of the program to the specified file
*/
void Credits( FILE *where)
{
	fprintf( where, "DMMPST: DooM MaP StaTistics, ver %s\n", DMMPST_VERSION);
	fprintf( where, " By Frans P. de Vries <fpv@gamers.org>\n");
/*	fprintf( where, "[Derived from DEU v%s by Brendon Wyber and Raphaël Quinet]\n\n", DEU_VERSION);
	fprintf( where, "Derived from DEU: Doom Editor Utility, ver %s\n\n", DEU_VERSION);
	fprintf( where, " By Raphaël Quinet (quinet@montefiore.ulg.ac.be)\n"
	                "and Brendon J Wyber (b.wyber@csc.canterbury.ac.nz)\n"
	                " Ported to DJGPP/GO32 by Per Allansson (c91peral@und.ida.liu.se)\n"
	                "and Per Kofod (per@ollie.dnk.hp.com)\n\n"); */
}


/*
	terminate the program reporting an error
*/
void ProgError( char *errstr, ...)
{
	va_list args;

	va_start( args, errstr);
	printf( "\nProgram Error: *** ");
	vprintf( errstr, args);
	printf( " ***\n");
	va_end( args);
	/* clean up & free space */
	ForgetLevelData();
	CloseWadFiles();
	exit( 1);
}


/*
	the main program "loop"
*/
void MainLoop()
{
	CheckLevelParams( Episode, Mission);

	if (WkFlName != NULL)
	{
		/* check for Doom II, Strife or Hexen */
		if (GameVersion == 0x02 || GameVersion >= 0x20)
			printf( "\nOutputting DoomWiki template of level MAP%02d to \"%s\".\n",
			        Mission, WkFlName);
		else
			printf( "\nOutputting DoomWiki template of level E%dM%d to \"%s\".\n",
			        Episode, Mission, WkFlName);
		if ((WkFile = fopen( WkFlName, "wt")) == NULL)
			ProgError( "error opening output file \"%s\"", WkFlName);
	}
	AnalyzeLevel( Episode, Mission);
	if (WkFlName != NULL)
		fclose( WkFile);
}


/*
	convert string to uppercase
*/
char *strupr( char *str)
{
	char *s = str;

	if (str != NULL)
		while (*s)
		{
			if (islower( *s))
				*s = toupper( *s);
			s++;
		}

	return str;
}

/* vim:set noexpandtab tabstop=2 softtabstop=2 shiftwidth=2: */
